/** use this in combination with sEnergy.ino **
 **********************************************/
 // This demonstration code switches a relay and LED according to
 //     when there's sunshine
 // with this project you can directly control for example a pond pump

//** START sWiFiFull.h definitions **/           // www.kendziorra.nl download: simpleLib
#define USE_STA      // use WiFi mode STA
#define USE_OTA      // ESP32 OTA
#define USE_TIME     // NTP, date, time, timezone, DST/Summertime, Sunrise, Sunset /** my simpleLIB library: sDateTimeSun.h **/

/** WiFi **/
const char *staSsid = "edit"@,
       *staPassword = "edit"@;
const IPAddress wsIP(192, 168, 178, 121);  // IP from Websocket server
const char *myOtaID = "P1_WS_2",   // OTA name
             *wsURI = "/",         // WebSocket URI
        *wsProtocol = "P1_ESP32";  // WebSocket Protocol
const uint8_t wsPort = 82;         // Websocket Port
const double latSun = 52.1234,@     // your location ...
             lonSun = 5.5678;@      //  ... used by sunrise/set
/** time **/
// ***** List of timezones https://sites.google.com/a/usapiens.com/opnode/time-zones
char timeZone[] = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // EU Amsterdam/Barcelona with Summertime (DST)
// char timeZone[] = "CET-1";  // no DST/Summertime - EU Amsterdam/Barcelona
const time_t syncTime = 12 * 3600;  // NTP time sync interval in sec. if 0 then NOT used

// after the simpleLib declarations!!!
#include <sWiFiFull.h>
#include <WebSocketsClient.h>  // WebSockets Client /** library: https://github.com/Links2004/arduinoWebSockets **/

String  wsBuf = "";      // we use a String so do not suffer from a character array length problem
bool wsVal = false;      // data received?
time_t t;                // local time (UNIX timestamp) from NTP servers
uint8_t relaysPin = 27,
           ledPin =  2;
bool relayOn = false,
       ledOn = false; 
const uint8_t priority = 0 ;        // if more than one client: then test for energy surplus in sequence
                                    // 0: direct, 1: after 5 minutes, 2: after 10 minutes, .... (only if energy surplus)
const double scope = 1.5,           // threshold scope
              watt = 0.045 / 12,    // requested power in kWh ( / 12: watt 5 minutes)
         threshold = watt * scope;  // switch point to use the ....

sRiseSet rs;  // Sun rs.rise, rs.set timestamps
WebSocketsClient webSocket;

void wsAction();  // message from a WebSocket received?
void relay(bool on);
void webSocketEvent(WStype_t type, uint8_t * payload, size_t length);

void setup() {
//Serial.begin(115200);  // only for testing
//while (!Serial) delay(100);
//Serial.println("\n");
  setupWiFi();  // is necessary on this place - init all your defined choices
	webSocket.begin(wsIP, wsPort, wsURI, wsProtocol);
  webSocket.onEvent(webSocketEvent);
	webSocket.setReconnectInterval(5000);
  pinMode(relaysPin, OUTPUT);
  digitalWrite(relaysPin, HIGH); // LOW: ON
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, LOW);
  t = now();
}

void loop() {
  static uint8_t saveMinute = 99;
  static time_t saveTime = t;
  static bool dl = false;  // sDaylight flag

  loopWiFi();
	webSocket.loop();
  t = now();
  if (minute(t) != saveMinute) {
    if ((hour(t) == 0 && minute(t) == 0) || saveMinute == 99) rs = sSun(t, latSun, lonSun);  // get local time sun rise/set
    dl = sDaylight(t, rs);
    saveMinute = minute(t);
  }
  if (dl) {
    if (wsVal) wsAction();  // read WebSocket
    if (relayOn && (t >= saveTime)) {
      digitalWrite(ledPin, ledOn);   
      ledOn = !ledOn;                
      saveTime = t + 1;  // blink, 1 sec. interval
    }
  }
  else if (relayOn) relay(false); 
}

void wsAction() {
  static int8_t count = -1;  // priority counter, update every 5 minutes
  double l = 0.0;  // SUN: L1 + L2 + L3 
  int8_t n = 0;

  wsVal = false;             
  if (wsBuf.startsWith("?")) {  // "?timeVal;sunVallue"
//Serial.printf("received: %s\n", wsBuf.c_str());
    /** get SUN overproduction value **/
    n = wsBuf.indexOf(";", n);  // value 1: sun overproduction
    n = wsBuf.indexOf(";", n + 1);  // value 2: sun production
    l = wsBuf.substring(n + 1).toDouble();  // sun shines?
    if (relayOn) l += watt; 
    if (l >= threshold) {  // using the SUN values?
      /** if using more then one SUN-user then set the priorities for each module **/
      count++;
//Serial.printf("count: %u\n", count);
      if (priority == count) {
        relay(true);
        count = -1;
      }
      if (count > priority) count = -1;     
    }
    else {
      relay(false);
      count = -1;           
    }
  }
}

void relay(bool on) {
  digitalWrite(relaysPin, !on);  // LOW: ON
  relayOn = on;
  if (!on) {
    digitalWrite(ledPin, LOW);
    ledOn = false;
  }
}

void webSocketEvent(WStype_t type, uint8_t * payload, size_t len) {
  switch(type) {
    case WStype_DISCONNECTED:  break;
    case WStype_CONNECTED: break;
    case WStype_TEXT: {
      if (len) {
        wsBuf = String((char*)payload);  // use this data
        wsVal = true;
      }
      break;
    }
  }
}
