//** START sWiFiFull.h definitions **/           // www.kendziorra.nl download: simpleLib
#define USE_STA      // use WiFi mode STA
#define USE_STAIP    // if using a static STA IP  (is needed with: USE_WS)
#define USE_STADNS   // use WiFi STA DNS (use it if USE_STAIP and USE_TIME)
#define USE_OTA      // ESP32 OTA
#define USE_TIME     // NTP, date, time, timezone, DST/Summertime, Sunrise, Sunset /** my simpleLIB library: sDateTimeSun.h **/
#define USE_WS       // WebSockets Server /** library: https://github.com/Links2004/arduinoWebSockets **/

// #################################################### //
//   Deze code is alleen op een DSMR 5.0 meter getest   //
//  Volledig vrij te gebruiken en te wijzigen, geen (c) //
//       Ewald Kendziorra sr - www.kendziorra.nl        //
// #################################################### //

const uint32_t baudRate    = 115200,@      // 9600       - before DSRM v4.0
               bitsParStop = SERIAL_8N1;@  // SERIAL_7E1 -   ''      '' 
const bool     inverted    = true;        // standard inverted
const uint8_t rxPin = 16,  // standard ESP32 ...
              txPin = -1;  // ... 'serial 2' pins
/** WiFi **/
const char *staSsid = "edit-edit",@
       *staPassword = "edit-edit";@
const IPAddress staIP(192, 168, 178, 121),@  // IP also in HTML
         staGatewayIP(192, 168, 178,   1),@
          staSubnetIP(255, 255, 255,   0);
const IPAddress staPrimDNS(8, 8, 8, 8),
                 staSecDNS(8, 8, 4, 4);
const char *myOtaID = "P1_ESP32",  // OTA name: P1_ESP32 fixed value
             *wsURI = "/",         // WebSocket URI,       also in HTML
        *wsProtocol = "P1_ESP32";  // WebSocket Protocol,  also in HTML
const uint8_t wsPort = 82,         // Websocket Port,      also in HTML
              svPort = 80,         // webserver port
              ssPin = 5;           // SD SS pin
const double latSun = 52.1234,@     // your location ...
             lonSun = 5.5678;@      //  ... used by sunrise/set
/** time **/
// ***** List of timezones https://sites.google.com/a/usapiens.com/opnode/time-zones
char timeZone[] = @"CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // EU Amsterdam/Barcelona with Summertime (DST)
// char timeZone[] = "CET-1";  // no DST/Summertime - EU Amsterdam/Barcelona
const time_t syncTime = 12 * 3600;  // NTP time sync interval in sec. if 0 then NOT used
// after the simpleLib declarations!!!
#include <sWiFiFull.h>
#include <WebServer.h>
#include <SD.h> // format the SD card on a suitable device (FAT16 and FAT32 are suported) 
#include <EEPROM.h>  // to store SD year
/**  P1 meter **/
const uint8_t OBIS_1_len = 5,  // the number of items in 'char OBIS_1/2'
              OBIS_2_len = 6,
             OBIS_gasPos = 4;  // position of "0-1:24.2.1" in OBIS_1 array
const char OBIS_1[OBIS_1_len][11] {  // only the most important information for use with solar panels 
            "1-0:1.8.1",  // low tarif    - Meter Reading electricity delivered to client in 0.001 kWh 
            "1-0:1.8.2",  // normal tarif -    saved and send together as one value          
            "1-0:2.8.1",  // low tarif    - Meter Reading electricity delivered by client in 0.001 kWh
            "1-0:2.8.2",  // normal tarif -    saved and send together as one value          
            "0-1:24.2.1"  // [Gas] Last 5-minute Meter reading in m3, 
           };  // five minute interval
const char OBIS_2[OBIS_2_len][11] {  // only the most important information for use with solar panels 
            //  "1-0:1.7.0" and "1-0:2.7.0",  // are the actual eL1+L2+L3 ectricity power delivered (+P/-P) in 1 Watt resolution 
            //     these 2 are the totals of the following 6 values L1 to L3, so we calculate this
            "1-0:21.7.0", // L1 Actual electricity power delivered (+P) in 1 Watt resolution
            "1-0:41.7.0", // L2   from network
            "1-0:61.7.0", // L3 
            "1-0:22.7.0", // L1 Actual electricity power received (-P) in 1 Watt resolution
            "1-0:42.7.0", // L2   to network
            "1-0:62.7.0", // L3 
           };  // one second interval
/** read and save values **/
const uint8_t requestPin = 27;    // the P1 send ON/OFF requestPin is used to avoid overflow in P1 buffer
const String svc = ";",           // svc fields seperation (if changed all the clients use this)
         logPath = "/",           // (used by "sDownloadLogFile.h") path to log file (at least one '/') DO NOT CHANGE
         logExt  = ".csv";        // log file extension
uint8_t logYearRom = 0;           // used for logfile year from EEPROM
bool sdPresent = false,           // SD-card present
          toWs = false;           // continuous sent to WebSocket
String logYear[2] = {"", ""},     // logfiles year value
         telegram = "",           // P1 telegram buffer
       totalsName = "",           // year-month totals file name
     downloadName = "";           // (used by Download LogFiles, file name [year]M[onth]D[ay], 21M11D12 = 12 Nov '21
                                  //    on SD-card: one day with 5 min. intervals 
time_t t;                         // local time (UNIX timestamp) from NTP servers
/** S0 pulsses **/
const uint16_t pulseRate = 1000;  // number of pulses/KWh, 1000: one pulse is 1.0W, 2000: one pulse is 0.5W
const uint8_t s0Pin = 14;         // interupt pin 
volatile uint16_t pulses = 0;     // pulses couter
double kwh = 0,                   // KWh value
    kwhOld = 0;
/** totals **/
double totals[12][OBIS_1_len + 1] = {0.0};  // 12x6 month / totals
                                
File logFile,                    // logFile object
     totalsFile;                  // year,month,day totals object
WebServer server(svPort);         // WebServer object
sRiseSet rs;                      // Sun rs.rise, rs.set timestamps

void IRAM_ATTR onPulse();         // get the S0 pulses
bool readP1();                    // serial read a telegram
bool getValues();                 // true: send result to WebSocket Client
void openLog();                   // open or create a day-logFile
bool yearTest();                  // test log years
void delFiles();                  // delete old log files
void downloadLog();               // download a day-logFile
void wsAction();                  // message from a WebSocket received?

void setup() {
  // only used if using USB-port to upload firmware or testing
//    Serial.begin(115200);
//    while (!Serial); delay(100);
//    Serial.println("\n");
  ////////////////////////////////////////////////////////////
  EEPROM.begin(1);            // logYearRom = 1 byte
  Serial2.begin(baudRate, bitsParStop, rxPin, txPin, inverted);  // P1 serial port
  pinMode(requestPin, OUTPUT);
  pinMode(s0Pin, INPUT);
  attachInterrupt(s0Pin, onPulse, FALLING);
  setupWiFi();  // is necessary on this place - init all your defined choices
  t = now();
  yearTest();
  sdPresent = SD.begin(ssPin);   // if TRUE: then you can save log values and download log files
  if (sdPresent) {               // else   : you can only use a WebSocket Client for real time information (webbrowser, ESP-module, or ...)
    server.on("/download", downloadLog);
    server.begin();
    totalsName = "totals" + logYear[0];  // year-month totals file
//////////////////////////////////////////
/** init with old values for this year **/
//////////////////////////////////////////
//      totalsFile = SD.open(logPath + totalsName + logExt, FILE_WRITE);
//                             //  inL, inH, sonL, sonH, gas, panels
//      double startVal[12][6] = {{170.0,143.0,0.0,0.0,264.0,0.0}, 
//                                {148.0,127.0,0.0,0.0,221.0,0.0},
//                                {183.376,204.923,10.224,29.199,146.810,98.007},  // 17 feb 14:00 start PV-panels
//                                {50.523,48.245,2.461,0.521,37.653,21.731},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0},
//                                {0.0,0.0,0.0,0.0,0.0,0.0}}; 
//      String s = "";
//      for (uint8_t i = 0; i < 12; i++) {
//        for (uint8_t ii = 0; ii <= OBIS_1_len; ii++) {
//          totals[i][ii] = startVal[i][ii];
//          s += (ii ? svc : "") + String(startVal[i][ii], 3);
//        }
//        s += '\n';
//      }
//      totalsFile.print(s);
//      totalsFile.close();
//////////////////////////////////////////
/****************************************/
//////////////////////////////////////////

    uint8_t m = 0;    // [month] 0 to 11
    uint8_t v = 0;    // [value] 0 to 5
    char c = '\0';    // read byte
    String val = "";  // String input buffer to float convert
    totalsFile = SD.open(logPath + totalsName + logExt);
    while (totalsFile.available()) {  // read total vallues [month 1-12][val 1-6]
      c = (char)totalsFile.read();     
      if (c == '\n') {
        if (val.length() > 2) totals[m][v++] = val.toDouble();  // convert total values string to double
        if (++m == 12) break;  // month: 0 to 11 
        v = 0;
        val = "";
      }
      else if (c == ';') {
        if (val.length() > 2) totals[m][v++] = val.toDouble();  // convert total values string to double
        val = "";
      }
      else val += c;
    }
    totalsFile.close(); 
    openLog();  // open log file
  }
}

void loop() {
  static uint8_t saveSecond = 99;    // new minute? save on SD - new day/read telegram
  loopWiFi();
  if (wsVal) wsAction();  // read WebSocket 
  if (sdPresent) {
    server.handleClient(); // Listen for client connections
    delay(2);
  }
  t = now();
  if (wsClients == 0) toWs = false;  // WebSocket clients active? (wsClients: from simpleLIB sWiFiFull.h)
  if (second(t) != saveSecond) {
    if (pulses) {
      kwh += (double)pulses / pulseRate;
      pulses = 0;
    }
    if (readP1()) {
      if (getValues()) saveSecond = second(t);
    }
    if (downloadName != (logYear[0] + "M" + String(month(t)) + "D" + String(day(t)))) {
      kwh = 0;
      kwhOld = 0;
      openLog();  // new day, new file?
    }
  }
}

void IRAM_ATTR onPulse() {
  static time_t ms = millis();
  if (millis() >= ms) {
    pulses++;
    ms = millis() + 100;
  }
}

bool readP1() {
  char c;
  bool more = false,  // start detected
       last = false;  // end detected
  String buf = "";
  time_t s = now() + 2;  // now + n seconds for max while loop
  digitalWrite(requestPin, HIGH);   // P1 start sending
  delay(2);
  while (!last && (now() < s)) {
    while(Serial2.available()) {
      c = Serial2.read();
      if (more) {
        if (c == '/') buf = "";     // restart reading
        if (c == '!') last = true;  // telegram end
        buf += c;  // add a value
      }
      else if (c == '/') {
        buf = c;
        more = true;                // start reading
      }
    }
  }
  digitalWrite(requestPin, LOW);    // P1 stop sending
  if (last) telegram = buf;
  return last;
}
  
bool getValues() {
  static uint8_t saveMinuut = 99;
  static double energy_1a[OBIS_1_len];  // last read totals
  double energy_1b[OBIS_1_len],  // totals: low_net, high_net, low_sun, high_sun, gas
         energy_2[OBIS_2_len],   // actual L1,L2,L3_net, L1,L2,L3_sun 
         helper;
  String result = "";
  int16_t p1;
  char buf[10];   
  
  sprintf(buf, "%02u:%02u:%02u", hour(t), minute(t), second(t));
  String timeVal = String(buf);
  if (saveMinuut == 99 || ((saveMinuut != minute(t) && (minute(t) % 5) == 0) && second(t) == 0)) {
    for (uint8_t i = 0; i < OBIS_1_len; i++) {
      p1 = telegram.indexOf(OBIS_1[i]);
      if (p1) {
        p1 += strlen(OBIS_1[i]) + 1;
        if (i == OBIS_gasPos) p1 = telegram.indexOf(')', p1) + 2;  // skip the GAS timestamp?
        energy_1b[i] = telegram.substring(p1).toDouble();  // set new values
      }
      else return false;
    }
    result = timeVal;
    double surplus = 0.0;
    for (uint8_t i = 0; i < OBIS_1_len; i++) {
      helper = energy_1b[i] - energy_1a[i];
      if (i == 2 || i == 3) surplus += helper;
      result = result + svc + String(helper, 3);
      if (saveMinuut != 99) totals[month(t) - 1][i] += helper;
    }
    helper = kwh - kwhOld;
    result = result + svc + String(helper, 3) + "\n";
    if (saveMinuut != 99) totals[month(t) - 1][OBIS_1_len] += helper;
    kwhOld = kwh;
    if (saveMinuut != 99){
      if (toWs) webSocket.broadcastTXT(">" + result);  // for standard WebSocket use
      if (sDaylight(t, rs)) webSocket.broadcastTXT("?" + timeVal + svc + String(surplus, 3) + svc + String(helper, 3)+ "\n");  // sun overproduction  and sun production - used by WebSocket Clients
      if (sdPresent) {
        logFile.print(result);
        logFile.flush();
        yield();
        String s = "";
        for (uint8_t i = 0; i < 12; i++) {
          yield();
          for (uint8_t ii = 0; ii <= OBIS_1_len; ii++) s += (ii ? svc : "") + String(totals[i][ii], 3);
          s += '\n';
        }
        totalsFile.print(s);
        totalsFile.flush();
      }
    }
    memcpy(energy_1a, energy_1b, sizeof(double)*OBIS_1_len);  // save previous values
    saveMinuut = minute(t);
  }
  for (uint8_t i = 0; i < OBIS_2_len; i++) {
    p1 = telegram.indexOf(OBIS_2[i]);
    if (p1) {
      p1 += strlen(OBIS_2[i]) + 1;
      energy_2[i] = telegram.substring(p1).toDouble();
    }
    else return false;
  }
  if (toWs) webSocket.broadcastTXT("#" + timeVal + svc + String(energy_2[0], 3) + svc + String(energy_2[1], 3) + svc + String(energy_2[2], 3) + svc +  String(energy_2[3], 3) + svc + String(energy_2[4], 3) + svc + String(energy_2[5], 3) + "\n");  // for standard WebSocket use
  return true;
}

void openLog() {
  bool ex = false;
  rs = sSun(t, latSun, lonSun);  // get local time sun rise/set
  String oldLog = downloadName;
  String oldTot = totalsName;
  downloadName = logYear[0] + "M" + String(month(t)) + "D" + String(day(t));
  totalsName = "totals" + logYear[0];
  if (oldLog != downloadName) {  // a new day?
    if (sdPresent) logFile.close();
    if (yearTest()) {     
      downloadName = logYear[0] + "M" + String(month(t)) + "D" + String(day(t));
      totalsName = "totals" + logYear[0];
      for (uint8_t i = 0; i < 12; i++) {
        for (uint8_t ii = 0; ii <= OBIS_1_len; ii++) totals[i][ii] = 0.0;
      }
      if (sdPresent) {
        if (oldTot != totalsName) {
          totalsFile.close();
          totalsFile = SD.open(logPath + totalsName + logExt, FILE_WRITE);
          for (uint8_t i = 0; i < 12; i++) totalsFile.print("0.0;0.0;0.0;0.0;0.0;0.0\n");
          totalsFile.flush();
        }
      }
    }
  }
  if (sdPresent) {
    ex = SD.exists(logPath + downloadName + logExt);
    logFile = SD.open(logPath + downloadName + logExt, FILE_APPEND);  // (re)open (new) log
    if (!ex) {  // column labels on first row 
     logFile.print("tijd;NET laag;NET hoog;ZON laag;ZON hoog;GAS m3;Productie\n");
     logFile.flush();
    }
  }
}

bool yearTest() {
  bool f = false;
  t = now();
  uint8_t y = year(t) - 2000;
  EEPROM.get(0, logYearRom);  // get the last used year value
  logYear[0] = String(y);
  logYear[1] = String(y-1);
  if (y != logYearRom) {
    EEPROM.put(0, y);
    EEPROM.commit();  // save year value
    logYearRom = y;
    delFiles();
    f = true;
  }
  return f;
}
  
void delFiles() {  // delete old files (this year - 2)
  if (!sdPresent) return;
  File root = SD.open("/");
  String y = "";
  while (true) {
    File entry = root.openNextFile();
    if (!entry) break;
    y = entry.name();
    if (!y.startsWith("totals")) {
      y = y.substring(0, 2);
      if (y != logYear[0] && y != logYear[1])  SD.remove(logPath + entry.name());
    }
    entry.close();
  }
  t = now();
}
  
void downloadLog(){
  String down;
  bool t = false;
  if (server.arg(0).startsWith("totals")) {
    down = server.arg(1);
    t = true;
  }
  else down  = (server.arg(1) == "0" ? logYear[0] : logYear[1]) + server.arg(0);
  if (!SD.exists(logPath + down + logExt)) {
    if (t) down = totalsName;
    else down = downloadName;
  }
  if (SD.exists(logPath + down + logExt)) {
    File download = SD.open(logPath + down + logExt);
    if (download) {
      server.sendHeader("Content-Type", "text/text");
      server.sendHeader("Content-Disposition", "attachment; filename=" + down + logExt);
      server.sendHeader("Connection", "close");
      server.streamFile(download, "application/octet-stream");
      download.close();
    }
  }
}

void wsAction() {
  wsVal = false;
  if (wsBuf == "start") toWs = true;
  if (wsBuf == "stop") toWs = false;
}
